#include <string.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

typedef ptrdiff_t Int;

typedef struct {
    bool main;
    bool off;
} parts;

typedef struct {
    char *s;
    parts *consumed;
    Int width;
    Int height;
} grid;

grid parsegrid(char *s) {
    grid g;
    g.s = s;
    g.width = strcspn(s, "\n");
    g.height = strlen(s) / g.width;
    // freed when the program dies
    g.consumed = calloc(g.width * g.height, sizeof * g.consumed);
    return g;
}

char indexgrid(grid g, Int x, Int y) {
    // width + 1 for the newlines
    return g.s[y*(g.width+1)+x];
}

parts *indexconsumed(grid g, Int x, Int y) {
    return &g.consumed[y*g.width+x];
}

bool inbounds(grid g, Int x, Int y) {
    return x >= 0 && y >= 0 && x < g.width && y < g.height;
}

bool followline(grid g, Int x, Int y, Int dx, Int dy) {
    Int ox = x;
    Int oy = y;

    while (inbounds(g, x, y)) {
        Int mainaltx = 0;
        Int mainalty = -1;

        switch (indexgrid(g, x, y)) {
            case '/': {
                Int olddy = dy;
                dy = -dx;
                dx = -olddy;
                mainaltx = 1;
                mainalty = 0;
            } case '\\': {
                Int olddy = dy;
                dy = dx;
                dx = olddy;
                mainaltx = -1;
                mainalty = 0;
            }
        }

        parts *p = indexconsumed(g, x, y);
        if (dy == 1 || (dx == mainaltx && dy == mainalty))
            p->main = true;
        else
            p->off = true;

        x += dx;
        y += dy;
    }

    return
        ((ox == 0 || ox == g.width-1) && (x == -1 || x == g.width) && llabs(ox-x) >= 8) ||
        ((oy == 0 || oy == g.height-1) && (y == -1 || y == g.height) && llabs(oy-y) >= 8);
}

bool hasconcluded(char *s) {
    grid g = parsegrid(s);

    #define follow(g, x, y, dx, dy) do { if (followline(g, x, y, dx, dy)) return true; } while (0)
    for (Int x = 0; x < g.width; x++) {
        follow(g, x, 0, 0, 1);
        follow(g, x, g.height-1, 0, -1);
    }
    for (Int y = 0; y < g.width; y++) {
        follow(g, 0, y, 1, 0);
        follow(g, g.width-1, y, -1, 0);
    }

    // loops are that left over
    for (Int y = 0; y < g.height; y++) {
        for (Int x = 0; x < g.width; x++) {
            parts p = *indexconsumed(g, x, y);
            if (!p.main || !p.off) return true;
        }
    }

    return false;
}
